#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <dbus/dbus.h>
#include <sane/sane.h>

//number of scans that can be open at the same time -
// 10 is sort of overkill, but it's better to be safe
#define NUM_SIMULT_SCANS 10
//name of output file for scanned image
#define OUT_FILENAME "scanned_img"
//size of buffer for scanned image
#define BUF_SIZE 1000

//TODO: make it exit main loop on SIGINT signal

void method_get_device_list (DBusMessage*, DBusConnection*);
void method_start_scan (DBusMessage*, DBusConnection*);
void method_stop_scan (DBusMessage*, DBusConnection*);
void method_read_image (DBusMessage*, DBusConnection*);

//table of dbus methods with functions to handle them
void *methodtable[] =
	{"GetDeviceList", method_get_device_list, "StartScan", method_start_scan,
	"StopScan", method_stop_scan, "ReadImage", method_read_image, NULL};
//stores handles for active scans
SANE_Handle scanhandles[NUM_SIMULT_SCANS];
//used to stop main loop on signal
int run = 1;

void send_reply_failure (DBusMessage *msg, DBusConnection *conn, int errcode) {
	DBusMessage *reply;
	DBusMessageIter args;
	int interrcode;

	interrcode = errcode;
	//create a reply message
	reply = dbus_message_new_method_return(msg);
	//add the error code to it
	dbus_message_iter_init_append(reply, &args);
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &errcode);
	//and send the reply
	dbus_connection_send(conn, reply, NULL);
	dbus_connection_flush(conn);
	dbus_message_unref(reply);
}

void method_get_device_list (DBusMessage *msg, DBusConnection *conn) {
	DBusMessage* reply;
	DBusMessageIter args;
	const SANE_Device **devicelist;
	SANE_Status status;
	int currdevice;

	//doesn't take any args, so we're not even going to check
	//get the device list from SANE
	status = sane_get_devices(&devicelist, FALSE);
	if (status != SANE_STATUS_GOOD) {
		send_reply_failure(msg, conn, status);
		return;
	}

	//create the reply
	reply = dbus_message_new_method_return(msg);
	//init the args for the reply
	dbus_message_iter_init_append(reply, &args);
	//first, the required status
	status = SANE_STATUS_GOOD;
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &status);
	/* now go through the list of devices, and add each one's name to the reply.
		an int ID for each would be better, but for now it's fine since the name
		is supposed to be unique */
	for (currdevice = 0; devicelist[currdevice] != NULL; currdevice++) {
		dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING,
			devicelist[currdevice]->name);
		printf("Device: %s\n", devicelist[currdevice]->name);
	}
	//now send the reply
	dbus_connection_send(conn, reply, NULL);
	dbus_connection_flush(conn);
	dbus_message_unref(reply);
}

void method_start_scan (DBusMessage *msg, DBusConnection *conn) {
	DBusMessage* reply;
	DBusMessageIter args;
	SANE_Status status;
	SANE_Handle devicehandle;
	char *devicename;

	//send a 'failed' return code if it didn't have any args
	if (!dbus_message_iter_init(msg, &args)) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL); 
		return;
	}
	//same if it isn't a string
	if (DBUS_TYPE_STRING != dbus_message_iter_get_arg_type(&args)) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL);
		return;
	}

	//otherwise, get the device name from the argument
	dbus_message_iter_get_basic(&args, &devicename);

	//and try to open the device
	status = sane_open(devicename, &devicehandle);
	if (status != SANE_STATUS_GOOD) {
		send_reply_failure(msg, conn, status);
		return;
	}

	//now start the scan
	status = sane_start(devicehandle);
	if (status != SANE_STATUS_GOOD) {
		sane_close(devicehandle);
		send_reply_failure(msg, conn, status);
		return;
	}

	//store the scan handle in our SANE_Handle table
	//and pass back an int to the caller to keep track of it
	int i = 0;
	while (scanhandles[i] != NULL) {
		i++;
		//prevent overflow if the table's already full
		if (i == NUM_SIMULT_SCANS) {
			send_reply_failure(msg, conn, SANE_STATUS_DEVICE_BUSY);
			return;
		}
	}

	//once we've found an empty slot, store the handle there
	scanhandles[i] = devicehandle;
	//and reply with a "success" status code
	reply = dbus_message_new_method_return(msg);
	dbus_message_iter_init_append(reply, &args);
	status = SANE_STATUS_GOOD;
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &status);
	//and the int
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &i);
	dbus_connection_send(conn, reply, NULL);
	dbus_connection_flush(conn);
	dbus_message_unref(reply);
}

void method_stop_scan (DBusMessage *msg, DBusConnection *conn) {
	DBusMessage* reply;
	DBusMessageIter args;
	SANE_Status status;
	int extscanhandle;

	//make sure it's passing a scan handle
	if (!dbus_message_iter_init(msg, &args)) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL); 
		return;
	}
	if (DBUS_TYPE_UINT32 != dbus_message_iter_get_arg_type(&args)) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL);
		return;
	}

	dbus_message_iter_get_basic(&args, &extscanhandle);
	//gotta make sure the scan handle's valid...
	if (extscanhandle >= NUM_SIMULT_SCANS || scanhandles[extscanhandle] == NULL) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL);
		return;
	}

	//finally, cancel the scan
	sane_close(scanhandles[extscanhandle]);	//what's the difference between this and sane_cancel?
	scanhandles[extscanhandle] = NULL;

	//and send back the status code
	reply = dbus_message_new_method_return(msg);
	dbus_message_iter_init_append(reply, &args);
	status = SANE_STATUS_GOOD;
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &status);
	dbus_connection_send(conn, reply, NULL);
	dbus_connection_flush(conn);
	dbus_message_unref(reply);
}

void method_read_image (DBusMessage *msg, DBusConnection *conn) {
	/* NOTE: this is just a temporary routine to test the image capture:
		the final version will have a much faster/more flexible way
		for passing back the image.
		for now, be careful since the file gets overwritten with
		each call to ReadImage, so two clients accessing this at the
		same time will probably start stepping on each other's toes. */

	DBusMessage* reply;
	DBusMessageIter args;
	SANE_Status status;
	int extscanhandle;
	SANE_Handle intscanhandle;
	char *tmpdir;
	FILE *tmpfile;
	int datalen;
	char databuf[BUF_SIZE];

	//make sure it's passing a scan handle
	if (!dbus_message_iter_init(msg, &args)) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL); 
		return;
	}
	if (DBUS_TYPE_UINT32 != dbus_message_iter_get_arg_type(&args)) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL);
		return;
	}

	dbus_message_iter_get_basic(&args, &extscanhandle);
	//the scan handle also has to be valid
	if (extscanhandle >= NUM_SIMULT_SCANS || scanhandles[extscanhandle] == NULL) {
		send_reply_failure(msg, conn, SANE_STATUS_INVAL);
		return;
	}

	//retrieve our SANE scan handle
	intscanhandle = scanhandles[extscanhandle];

	/* now, we're going to read in the image and dump
		the raw data to a temporary file, then pass back
		its name to the calling program */

	//first, find a temp dir
	tmpdir = getenv("TMPDIR");
	if (!tmpdir)
		tmpdir = "/tmp";
	if (tmpdir[strlen(tmpdir) - 1] == '/')
		tmpdir[strlen(tmpdir) - 1] = '\0';
	//then open the file
	char tmppath[strlen(tmpdir) + strlen(OUT_FILENAME) + 1];
	strcpy(tmppath, tmpdir);
	strcat(tmppath, "/");
	strcat(tmppath, OUT_FILENAME);

	tmpfile = fopen(tmppath, "w");
	if (!tmpfile) {
		send_reply_failure(msg, conn, SANE_STATUS_NO_MEM);
		return;
	}

	//now read from the device until the end of the image
	status = SANE_STATUS_GOOD;
	while (status == SANE_STATUS_GOOD) {
		status = sane_read(intscanhandle, databuf, BUF_SIZE, &datalen);
		fwrite(databuf, 1, datalen, tmpfile);
	}
	//TODO: have it time out if it doesn't read any data for a certain amount of time

	//finally, close the file
	fclose(tmpfile);

	//and send back the status code w/the output file's path
	reply = dbus_message_new_method_return(msg);
	dbus_message_iter_init_append(reply, &args);
	//if the return code was EOF, transfer was successful, so set status to show that
	if (status == SANE_STATUS_EOF)
		status = SANE_STATUS_GOOD;
	dbus_message_iter_append_basic(&args, DBUS_TYPE_UINT32, &status);
	dbus_message_iter_append_basic(&args, DBUS_TYPE_STRING, tmppath);
	dbus_connection_send(conn, reply, NULL);
	dbus_connection_flush(conn);
	dbus_message_unref(reply);
}


void handle_signal (int signal) {
	//if we caught a SIGINT (Ctrl-C), stop the main loop
	printf("Exiting...\n");
	run = 0;
}


DBusHandlerResult handle_request (DBusConnection *conn, DBusMessage *msg, void *userdata) {

	//look through the method table to see if we can handle this message
	int i = 0;
	while (TRUE) {
		//make sure we're not at the end of the table
		if (methodtable[i] == NULL) {
			printf("Got unrecognized method call\n");
			return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
		//if not, check if this call matches the received one
		} else if (dbus_message_is_method_call(msg,
			"org.freedesktop.ImageCaptureInterface", methodtable[i])) {

			//if it does, call the appropriate method
			printf("Method %s called\n", (char *)methodtable[i]);
			void (*callback)(DBusMessage*, DBusConnection*) =
				(void (*)(DBusMessage*, DBusConnection*))methodtable[++i];
			//TODO: create new thread for function so long I/O operations don't block requests
			callback(msg, conn);
			break;
		}
		//if it doesn't match, look at the next one
		i += 2;
	}
	return DBUS_HANDLER_RESULT_HANDLED;
}


int main (int argc, char **argv) {

	DBusConnection *conn;
	DBusError error;

	//init SANE
	if (sane_init(NULL, NULL) != SANE_STATUS_GOOD) {
		printf("Couldn't init SANE!\n");
		exit(1);
	}

	dbus_error_init(&error);
	//connect to DBUS
	conn = dbus_bus_get(DBUS_BUS_SESSION, &error);
	if (!conn) {
		printf("Couldn't connect to DBUS!\n");
		if (dbus_error_is_set(&error)) {
			printf("%s\n", error.message);
			dbus_error_free(&error);
		}
		sane_exit();
		exit(1);
	}

	//try to get a name
	if (dbus_bus_request_name(conn, "org.freedesktop.ImageCapture",
		DBUS_NAME_FLAG_REPLACE_EXISTING, &error) !=
		DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {

		printf("Couldn't lock name!\n");
		if (dbus_error_is_set(&error)) {
			printf("%s\n", error.message);
			dbus_error_free(&error);
		}
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}

	//export an object for requests
	DBusObjectPathVTable vtable;
	vtable.message_function = handle_request;
	vtable.unregister_function = NULL;
	if (!dbus_connection_register_object_path(conn, "/org/freedesktop/ImageCapture",
		&vtable, NULL)) {

		printf("Couldn't export object!\n");
		dbus_connection_unref(conn);
		sane_exit();
		exit(1);
	}

	//install a signal handler so the daemon can exit cleanly,
	struct sigaction sighandler = {handle_signal};
	if (sigaction(SIGINT, &sighandler, NULL) == -1)
		printf("Couldn't install signal handler, continuing...\n");

	//and do nothing until we get an exit signal
	while (dbus_connection_read_write_dispatch(conn, 100) && run) { }
	//TODO: run timeout counter here, keep track of each open scan handle
	// and close it if no activity for X seconds

	dbus_connection_unref(conn);
	sane_exit();
	return 0;
}
